from __future__ import annotations

from typing import NewType, TypeAlias

# ItemID supports hierarchical IDs using "." separator for nested item structures
# Example: "order.123" or "order.123.item.456"
ItemID = NewType("ItemID", str)

# NodeID identifies a node in the graph (type alias for str to work with RetworkXStrDiGraph)
NodeID: TypeAlias = str



from __future__ import annotations

import orjson

from sera.models._property import GetSCPropValueFunc, SystemControlledAttrs


def parse_system_controlled_attrs(
    attrs: dict | None,
) -> SystemControlledAttrs | None:
    if attrs is None:
        return None
    if not isinstance(attrs, dict):
        raise NotImplementedError(attrs)

    if "on_upsert" in attrs:
        attrs = attrs.copy()
        attrs.update(
            {
                "on_create": attrs["on_upsert"],
                "on_create_bypass": attrs.get("on_upsert_bypass"),
                "on_update": attrs["on_upsert"],
                "on_update_bypass": attrs.get("on_upsert_bypass"),
            }
        )

    if "on_create" not in attrs or "on_update" not in attrs:
        raise ValueError(
            "System controlled attributes must have 'on_create', 'on_update', or 'on_upsert' must be defined."
        )

    if "on_search" in attrs:
        attrs.update(
            {
                "on_search": attrs["on_search"],
                "on_search_bypass": attrs.get("on_search_bypass"),
            }
        )

    keys = {}
    for key in ["on_create", "on_update", "on_search"]:
        if key not in attrs:
            continue

        if attrs[key] == "ignored":
            keys[key] = "ignored"
        elif attrs[key].find(":") != -1:
            func, args = attrs[key].split(":")
            assert func == "getattr", f"Unsupported function: {func}"
            args = orjson.loads(args)
            keys[key] = GetSCPropValueFunc(
                func=func,
                args=args,
            )
        else:
            raise ValueError(
                f"System controlled attribute '{key}' must be 'ignored' or a function call in the format '<funcname>:<args>'."
            )

        if attrs[key + "_bypass"] is not None:
            if not isinstance(attrs[key + "_bypass"], str):
                raise ValueError(
                    f"System controlled attribute '{key}_bypass' must be a string."
                )
            keys[key + "_bypass"] = attrs[key + "_bypass"]

    return SystemControlledAttrs(
        on_create=keys["on_create"],
        on_create_bypass=keys.get("on_create_bypass"),
        on_update=keys["on_update"],
        on_update_bypass=keys.get("on_update_bypass"),
        on_search=keys.get("on_search", "ignored"),
        on_search_bypass=keys.get("on_search_bypass"),
    )



from __future__ import annotations

from codegen.models import ImportHelper, PredefinedFn, Program, expr, stmt

from sera.make.py_backend.misc import get_python_property_name
from sera.make.ts_frontend.misc import TS_GLOBAL_IDENTS, get_normalizer
from sera.misc import assert_not_null, to_camel_case, to_pascal_case
from sera.models import (
    Class,
    DataProperty,
    Enum,
    ObjectProperty,
    Package,
    Schema,
    TsTypeWithDep,
)


def make_class_schema(schema: Schema, cls: Class, pkg: Package):
    """Make schema definition for the class in frontend so that components can use this information select
    appropriate components to display or edit the data.

    Args:
        schema: The overall schema of the application, which contains all classes & enums
        cls: The class that we want to generate the schema
        pkg: The output package (directory) for the class in the `@.models` package. For example, if the
            class is `User`, then the package would be `src/models/user`.

    Returns:
        This function do not return anything as it writes the schema directly to a file.
    """
    if not cls.is_public:
        # skip classes that are not public
        return

    program = Program()
    prop_defs: list[tuple[DataProperty | ObjectProperty, expr.Expr, expr.Expr]] = []

    import_helper = ImportHelper(program, TS_GLOBAL_IDENTS)

    for prop in cls.properties.values():
        # we must include private properties that are needed during upsert for our forms.
        # if prop.data.is_private:
        #     # skip private fields as this is for APIs exchange
        #     continue
        tspropname = to_camel_case(prop.name)
        pypropname = prop.name
        if isinstance(prop, ObjectProperty) and prop.target.db is not None:
            # this is a database object, we append id to the property name
            tspropname = tspropname + "Id"
            pypropname = get_python_property_name(prop)

        tsprop = {}

        if isinstance(prop, DataProperty):
            tstype = prop.get_data_model_datatype().get_typescript_type()
            # for schema definition, we need to use the original type, not the type alias
            # if prop.name == idprop.name:
            #     # use id type alias
            #     tstype = TsTypeWithDep(f"{cls.name}Id")
            for dep in tstype.deps:
                program.import_(dep, True)
            tsprop = [
                (
                    expr.ExprIdent("datatype"),
                    (
                        expr.ExprConstant(tstype.spectype)
                        if tstype.type not in schema.enums
                        else expr.ExprConstant("enum")
                    ),
                ),
                *(
                    [
                        (
                            expr.ExprIdent("enumType"),
                            export_enum_info(program, schema.enums[tstype.type]),
                        )
                    ]
                    if tstype.type in schema.enums
                    else []
                ),
                *(
                    [
                        (
                            expr.ExprIdent("foreignKeyTarget"),
                            expr.ExprConstant(prop.db.foreign_key.name),
                        )
                    ]
                    if prop.db is not None
                    and prop.db.is_primary_key
                    and prop.db.foreign_key is not None
                    else []
                ),
                *(
                    [
                        (
                            expr.ExprIdent("weakRef"),
                            PredefinedFn.dict(
                                [
                                    (
                                        expr.ExprIdent("targetClass"),
                                        expr.ExprConstant(
                                            prop.weakref.target_class.name
                                        ),
                                    ),
                                    (
                                        expr.ExprIdent("targetProp"),
                                        expr.ExprConstant(
                                            prop.weakref.target_prop.name
                                        ),
                                    ),
                                    (
                                        expr.ExprIdent("onDelete"),
                                        expr.ExprConstant(prop.weakref.on_delete.value),
                                    ),
                                    (
                                        expr.ExprIdent("onUpdate"),
                                        expr.ExprConstant(prop.weakref.on_update.value),
                                    ),
                                ]
                            ),
                        )
                    ]
                    if prop.weakref is not None
                    else []
                ),
                (
                    expr.ExprIdent("isRequired"),
                    expr.ExprConstant(
                        not prop.is_optional
                        and prop.default_value is None
                        and prop.default_factory is None
                    ),
                ),
            ]

            norm_func = get_normalizer(tstype, import_helper)
            if norm_func is not None:
                # we have a normalizer for this type
                tsprop.append(
                    (
                        expr.ExprIdent("normalizer"),
                        norm_func,
                    )
                )
        else:
            assert isinstance(prop, ObjectProperty)
            if prop.target.name != cls.name:
                program.import_(
                    f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.get_tsmodule_name()}-schema.{prop.target.name}Schema",
                    True,
                )
            if prop.target.db is not None:
                # this class is stored in the database, we store the id instead
                tstype = (
                    assert_not_null(prop.target.get_id_property())
                    .get_data_model_datatype()
                    .get_typescript_type()
                )
            else:
                # we are going to store the whole object
                tstype = TsTypeWithDep(
                    type=prop.target.name,
                    spectype=prop.target.name,
                    deps=[
                        f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.get_tsmodule_name()}.{prop.target.name}"
                    ],
                )

            # we don't store the type itself, but just the name of the type
            # so not need to import the dependency
            # if tstype.dep is not None:
            #     program.import_(
            #         tstype.dep,
            #         True,
            #     )

            tsprop = [
                (
                    expr.ExprIdent("targetClass"),
                    expr.ExprConstant(prop.target.name),
                ),
                (
                    expr.ExprIdent("datatype"),
                    expr.ExprConstant(
                        tstype.spectype if prop.target.db is not None else "undefined"
                    ),
                ),
                (
                    expr.ExprIdent("cardinality"),
                    expr.ExprConstant(prop.cardinality.value),
                ),
                (
                    expr.ExprIdent("isEmbedded"),
                    expr.ExprConstant(prop.target.db is None),
                ),
                (
                    expr.ExprIdent("isRequired"),
                    expr.ExprConstant(not prop.is_optional),
                ),
                (
                    expr.ExprIdent("property"),
                    expr.ExprRawTypescript(
                        "function (prop: keyof typeof "
                        + f"{prop.target.name}Schema.allProperties) {{ "
                        + f"return new {import_helper.use('NestedProperty').ident}(this, {prop.target.name}Schema.allProperties[prop]); "
                        + "}"
                    ),
                ),
            ]

        prop_defs.append(
            (
                prop,
                expr.ExprIdent(tspropname),
                PredefinedFn.dict(
                    [
                        (expr.ExprIdent("name"), expr.ExprConstant(pypropname)),
                        (expr.ExprIdent("tsName"), expr.ExprConstant(tspropname)),
                        (
                            expr.ExprIdent("updateFuncName"),
                            expr.ExprConstant(f"update{to_pascal_case(prop.name)}"),
                        ),
                        (
                            expr.ExprIdent("label"),
                            expr.ExprConstant(prop.label.to_dict()),
                        ),
                        (
                            expr.ExprIdent("description"),
                            (
                                expr.ExprConstant(prop.description.to_dict())
                                if not prop.description.is_empty()
                                else expr.ExprConstant("undefined")
                            ),
                        ),
                        (
                            expr.ExprIdent("constraints"),
                            PredefinedFn.list(
                                [
                                    expr.ExprConstant(
                                        constraint.get_typescript_constraint()
                                    )
                                    for constraint in prop.data.constraints
                                ]
                            ),
                        ),
                        (
                            expr.ExprIdent("validator"),
                            PredefinedFn.attr_getter(
                                expr.ExprIdent(f"draft{cls.name}Validators"),
                                expr.ExprIdent(tspropname),
                            ),
                        ),
                    ]
                    + tsprop
                ),
            )
        )

    for type in ["ObjectProperty", "DataProperty"]:
        program.import_(f"sera-db.{type}", True)
    if cls.db is not None:
        program.import_(f"sera-db.Schema", True)
    else:
        program.import_(f"sera-db.EmbeddedSchema", True)

    program.import_(
        f"@.models.{pkg.dir.name}.{cls.get_tsmodule_name()}.{cls.name}", True
    )
    program.import_(
        f"@.models.{pkg.dir.name}.draft-{cls.get_tsmodule_name()}.Draft{cls.name}", True
    )
    program.import_(
        f"@.models.{pkg.dir.name}.draft-{cls.get_tsmodule_name()}.draft{cls.name}Validators",
        True,
    )
    if cls.db is not None:
        program.import_(
            f"@.models.{pkg.dir.name}.{cls.get_tsmodule_name()}.{cls.name}Id", True
        )

    program.root(
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"type {cls.name}SchemaType = "
            + PredefinedFn.dict(
                (
                    [
                        (expr.ExprIdent("id"), expr.ExprIdent(f"{cls.name}Id")),
                    ]
                    if cls.db is not None
                    else []
                )
                + [
                    (
                        expr.ExprIdent("publicProperties"),
                        expr.ExprIdent(
                            " | ".join(
                                [
                                    expr.ExprConstant(
                                        to_camel_case(prop.name) + "Id"
                                        if isinstance(prop, ObjectProperty)
                                        and prop.target.db is not None
                                        else to_camel_case(prop.name)
                                    ).to_typescript()
                                    for prop in cls.properties.values()
                                    if not prop.data.is_private
                                ]
                            )
                        ),
                    ),
                    (
                        expr.ExprIdent("allProperties"),
                        expr.ExprIdent(
                            f"{cls.name}SchemaType['publicProperties']"
                            + (
                                " | "
                                + " | ".join(
                                    [
                                        expr.ExprConstant(
                                            to_camel_case(prop.name)
                                        ).to_typescript()
                                        for prop in cls.properties.values()
                                        if prop.data.is_private
                                    ]
                                )
                                if any(
                                    prop.data.is_private
                                    for prop in cls.properties.values()
                                )
                                else ""
                            )
                        ),
                    ),
                    (
                        expr.ExprIdent("cls"),
                        expr.ExprIdent(cls.name),
                    ),
                    (
                        expr.ExprIdent("draftCls"),
                        expr.ExprIdent(f"Draft{cls.name}"),
                    ),
                ]
            ).to_typescript()
            + ";",
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            "const publicProperties = "
            + PredefinedFn.dict(
                [
                    (prop_name, prop_def)
                    for prop, prop_name, prop_def in prop_defs
                    if not prop.data.is_private
                ]
            ).to_typescript()
            + f" satisfies Record<{cls.name}SchemaType['publicProperties'], DataProperty | ObjectProperty>;"
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export const {cls.name}Schema = "
            + PredefinedFn.dict(
                [
                    (
                        expr.ExprIdent("publicProperties"),
                        expr.ExprIdent("publicProperties"),
                    ),
                    (
                        expr.ExprIdent("allProperties"),
                        expr.ExprRawTypescript(
                            "{ ...publicProperties, "
                            + ", ".join(
                                [
                                    f"{prop_name.to_typescript()}: {prop_def.to_typescript()}"
                                    for prop, prop_name, prop_def in prop_defs
                                    if prop.data.is_private
                                ]
                            )
                            + f"}} satisfies Record<{cls.name}SchemaType['allProperties'], DataProperty | ObjectProperty>"
                        ),
                    ),
                ]
                + (
                    [
                        (
                            expr.ExprIdent("primaryKey"),
                            expr.ExprRawTypescript(
                                expr.ExprConstant(
                                    assert_not_null(cls.get_id_property()).name
                                ).to_typescript()
                                + " as const"
                            ),
                        )
                    ]
                    if cls.db is not None
                    else []
                )
            ).to_typescript()
            + ";"
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export const Typed{cls.name}Schema: Schema<{cls.name}SchemaType['id'], {cls.name}SchemaType['cls'], {cls.name}SchemaType['draftCls'], {cls.name}SchemaType['publicProperties'], {cls.name}SchemaType['allProperties'], {cls.name}SchemaType> = {cls.name}Schema;"
            if cls.db is not None
            else f"export const Typed{cls.name}Schema: EmbeddedSchema<{cls.name}SchemaType['cls'], {cls.name}SchemaType['draftCls'], {cls.name}SchemaType['publicProperties'], {cls.name}SchemaType['allProperties']> = {cls.name}Schema;"
        ),
    )
    pkg.module(cls.get_tsmodule_name() + "-schema").write(program)


def export_enum_info(program: Program, enum: Enum) -> expr.Expr:
    """Export enum information to

    ```
    {
        type: <EnumType>,
        label: { [value]: MultiLingualString },
        description: { [value]: MultiLingualString }
    }
    ```
    """
    for key in ["Label", "Description"]:
        program.import_(f"@.models.enums.{enum.name}{key}", True)

    return PredefinedFn.dict(
        [
            (expr.ExprIdent("type"), expr.ExprIdent(enum.name)),
            (expr.ExprIdent("label"), expr.ExprIdent(enum.name + "Label")),
            (expr.ExprIdent("description"), expr.ExprIdent(enum.name + "Description")),
        ]
    )



from __future__ import annotations

from codegen.models import DeferredVar, ImportHelper, PredefinedFn, Program, expr, stmt

from sera.make.py_backend.misc import get_python_property_name
from sera.make.py_backend.system_controlled_property import (
    AvailableVars,
    get_controlled_property_value,
)
from sera.misc import assert_not_null
from sera.models import DataCollection, Module, Package
from sera.typing import GLOBAL_IDENTS


def make_python_get_by_id_api(
    collection: DataCollection, target_pkg: Package
) -> tuple[Module, str]:
    """Make an endpoint for querying resource by id"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.get", True)
    program.import_("litestar.status_codes", True)
    program.import_("litestar.exceptions.HTTPException", True)
    program.import_(
        app.logic.path
        + f".{collection.get_pymodule_name()}.search.{collection.get_search_service_name()}",
        True,
    )
    program.import_(
        app.models.data.path + f".{collection.get_pymodule_name()}.{collection.name}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls

    # Find system_controlled properties that should be applied on search/get
    search_controlled_props = [
        prop
        for prop in collection.cls.properties.values()
        if prop.data.system_controlled is not None
        and prop.data.system_controlled.is_on_search_value_updated()
    ]
    if len(search_controlled_props) > 0:
        program.import_(
            "sera.libs.api_helper.SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY", True
        )
        program.import_("litestar.types.Scope", True)

    available_controlled_prop_vars: AvailableVars = {
        "user": PredefinedFn.item_getter(
            expr.ExprIdent("scope"),
            expr.ExprConstant("user"),
        )
    }

    id_prop = assert_not_null(cls.get_id_property())
    id_type = id_prop.datatype.get_python_type().type

    # Check if any controlled property is the ID property itself
    # In this case, we compare the user's value directly with the ID parameter
    is_id_controlled = any(
        prop.name == id_prop.name for prop in search_controlled_props
    )
    search_controlled_props = [
        prop for prop in search_controlled_props if prop.name != id_prop.name
    ]

    func_name = "get_by_id"
    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("get"),
                [
                    expr.ExprConstant("/{id:%s}" % id_type),
                ],
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "id",
                    expr.ExprIdent(id_type),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
                DeferredVar.simple("scope", expr.ExprIdent("Scope"))
                if len(search_controlled_props) > 0 or is_id_controlled
                else None,
            ],
            return_type=expr.ExprIdent("dict"),
            is_async=True,
        )(
            stmt.SingleExprStatement(expr.ExprConstant("Retrieving record by id")),
            lambda ast100: ast100.assign(
                DeferredVar.simple("service"),
                expr.ExprFuncCall(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent(collection.get_search_service_name()),
                        expr.ExprIdent("get_instance"),
                    ),
                    [],
                ),
            ),
            # Generate code for controlled properties that match the ID
            # These compare user's value directly with the id parameter
            *(
                [
                    stmt.LineBreak(),
                    lambda ast_if: ast_if.if_(
                        # if not scope["state"][SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY]:
                        expr.ExprNegation(
                            PredefinedFn.item_getter(
                                PredefinedFn.item_getter(
                                    expr.ExprIdent("scope"),
                                    expr.ExprConstant("state"),
                                ),
                                expr.ExprIdent(
                                    "SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY"
                                ),
                            )
                        )
                    )(
                        # if scope["user"].tenant_id != id:
                        lambda ast_inner_if: ast_inner_if.if_(
                            expr.ExprNotEqual(
                                get_controlled_property_value(
                                    assert_not_null(
                                        id_prop.data.system_controlled
                                    ).get_on_search_update_func(),
                                    available_controlled_prop_vars,
                                ),
                                expr.ExprIdent("id"),
                            )
                        )(
                            lambda ast_raise: ast_raise.raise_exception(
                                expr.StandardExceptionExpr(
                                    expr.ExprIdent("HTTPException"),
                                    [
                                        PredefinedFn.keyword_assignment(
                                            "status_code",
                                            expr.ExprIdent(
                                                "status_codes.HTTP_404_NOT_FOUND"
                                            ),
                                        ),
                                        PredefinedFn.keyword_assignment(
                                            "detail",
                                            expr.ExprIdent(
                                                'f"Record with id {id} not found"'
                                            ),
                                        ),
                                    ],
                                )
                            )
                        ),
                    ),
                    stmt.LineBreak(),
                ]
                if is_id_controlled
                else []
            ),
            # Generate if/else for controlled properties that don't match ID (e.g., tenant_id on User)
            *(
                [
                    stmt.LineBreak(),
                    lambda ast_if: ast_if.if_(
                        # if not (scope["state"][SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY]):
                        expr.ExprNegation(
                            PredefinedFn.item_getter(
                                PredefinedFn.item_getter(
                                    expr.ExprIdent("scope"),
                                    expr.ExprConstant("state"),
                                ),
                                expr.ExprIdent(
                                    "SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY"
                                ),
                            )
                        )
                    )(
                        # e.g., tenant_id = scope["user"].tenant_id
                        *[
                            lambda ast_then: ast_then.assign(
                                DeferredVar.simple(get_python_property_name(prop)),
                                get_controlled_property_value(
                                    assert_not_null(
                                        prop.data.system_controlled
                                    ).get_on_search_update_func(),
                                    available_controlled_prop_vars,
                                ),
                            )
                            for prop in search_controlled_props
                        ]
                    ),
                    lambda ast_else: ast_else.else_()(
                        # e.g., tenant_id = None
                        *[
                            lambda ast_else: ast_else.assign(
                                DeferredVar.simple(get_python_property_name(prop)),
                                expr.ExprConstant(None),
                            )
                            for prop in search_controlled_props
                        ]
                    ),
                    stmt.LineBreak(),
                ]
                if len(search_controlled_props) > 0
                else []
            ),
            lambda ast11: ast11.assign(
                DeferredVar.simple("record"),
                expr.ExprAwait(
                    expr.ExprFuncCall(
                        expr.ExprIdent("service.get_by_id"),
                        [
                            expr.ExprIdent("session"),
                            expr.ExprIdent("id"),
                        ]
                        + [
                            PredefinedFn.keyword_assignment(
                                get_python_property_name(prop),
                                expr.ExprIdent(get_python_property_name(prop)),
                            )
                            for prop in search_controlled_props
                        ],
                    )
                ),
            ),
            lambda ast12: ast12.if_(PredefinedFn.is_null(expr.ExprIdent("record")))(
                lambda ast23: ast23.raise_exception(
                    expr.StandardExceptionExpr(
                        expr.ExprIdent("HTTPException"),
                        [
                            PredefinedFn.keyword_assignment(
                                "status_code",
                                expr.ExprIdent("status_codes.HTTP_404_NOT_FOUND"),
                            ),
                            PredefinedFn.keyword_assignment(
                                "detail",
                                expr.ExprIdent('f"Record with id {id} not found"'),
                            ),
                        ],
                    )
                )
            ),
            lambda ast13: ast13.return_(
                PredefinedFn.dict(
                    [
                        (
                            PredefinedFn.attr_getter(
                                expr.ExprIdent(cls.name), expr.ExprIdent("__name__")
                            ),
                            PredefinedFn.list(
                                [
                                    expr.ExprFuncCall(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent(cls.name),
                                            expr.ExprIdent("from_db"),
                                        ),
                                        [expr.ExprIdent("record")],
                                    )
                                ]
                            ),
                        )
                    ]
                ),
            ),
        ),
    )

    outmod = target_pkg.module("get_by_id")
    outmod.write(program)

    return outmod, func_name



from __future__ import annotations

from typing import Optional

from codegen.models import ImportHelper, expr

from sera.models import TsTypeWithDep

TS_GLOBAL_IDENTS = {
    "normalizers.normalizeNumber": "sera-db.normalizers",
    "normalizers.normalizeOptionalNumber": "sera-db.normalizers",
    "normalizers.normalizeDate": "sera-db.normalizers",
    "normalizers.normalizeOptionalDate": "sera-db.normalizers",
    "NestedProperty": "sera-db.NestedProperty",
}


def get_normalizer(
    tstype: TsTypeWithDep, import_helper: ImportHelper
) -> Optional[expr.ExprIdent]:
    if tstype.type == "number":
        return import_helper.use("normalizers.normalizeNumber")
    if tstype.type == "number | undefined":
        return import_helper.use("normalizers.normalizeOptionalNumber")
    if tstype.type == "Date":
        return import_helper.use("normalizers.normalizeDate")
    if tstype.type == "Date | undefined":
        return import_helper.use("normalizers.normalizeOptionalDate")

    assert "number" not in tstype.type, tstype.type
    return None



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sera.libs.digraph._execution_summary import ExecutionSummary
from sera.libs.digraph._node_output import NodeOutput
from sera.libs.digraph._types import ItemID, NodeID

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession


class Context:
    """Read-only context available during node execution.

    Provides access to:
    - Global variables (e.g., database sessions)
    - Previous node outputs and execution summaries

    This is the base context class. Subclass for specific use cases
    (e.g., LitestarContext for Litestar API handlers).
    """

    def __init__(
        self,
        outputs: dict[tuple[NodeID, ItemID], NodeOutput[Any]] | None = None,
        summaries: dict[tuple[NodeID, ItemID], ExecutionSummary] | None = None,
    ):
        self._outputs = outputs or {}
        self._summaries = summaries or {}

    def get_output_exec_summary(
        self, node_id: NodeID, input_id: ItemID
    ) -> ExecutionSummary:
        """Get the execution summary from a previous node for a given input ID.

        Args:
            node_id: The ID of the node whose summary to retrieve.
            input_id: The ItemID of the input that was processed.

        Returns:
            The ExecutionSummary from the specified node for the given input.

        Raises:
            KeyError: If no summary exists for the given node_id and input_id.
        """
        return self._summaries[(node_id, input_id)]

    def get_output(self, node_id: NodeID, input_id: ItemID) -> NodeOutput[Any]:
        """Get the output from a previous node for a given input ID.

        Args:
            node_id: The ID of the node whose output to retrieve.
            input_id: The ItemID of the input that was processed.

        Returns:
            The NodeOutput from the specified node for the given input.

        Raises:
            KeyError: If no output exists for the given node_id and input_id.
        """
        return self._outputs[(node_id, input_id)]

    def _set_result(
        self,
        node_id: NodeID,
        input_id: ItemID,
        output: NodeOutput[Any],
        summary: ExecutionSummary,
    ) -> None:
        """Internal method to store a node's result.

        This should only be called by the Graph during execution.
        """
        self._outputs[(node_id, input_id)] = output
        self._summaries[(node_id, input_id)] = summary


class LitestarContext(Context):
    """Context for Litestar + SQLAlchemy API request handlers.

    This context is only valid during a Litestar backend API request.
    It provides access to the SQLAlchemy AsyncSession for database operations.
    """

    def __init__(
        self,
        session: AsyncSession,
        outputs: dict[tuple[NodeID, ItemID], NodeOutput[Any]] | None = None,
        summaries: dict[tuple[NodeID, ItemID], ExecutionSummary] | None = None,
    ):
        super().__init__(outputs, summaries)
        self.session = session



from __future__ import annotations

from codegen.models import DeferredVar, ImportHelper, PredefinedFn, Program, expr, stmt

from sera.make.py_backend.system_controlled_property import (
    AvailableVars,
    get_controlled_property_value,
)
from sera.misc import assert_not_null
from sera.models import DataCollection, Module, Package
from sera.typing import GLOBAL_IDENTS


def make_python_update_api(
    collection: DataCollection, target_pkg: Package
) -> tuple[Module, str]:
    """Make an endpoint for updating resource using the graph API"""
    app = target_pkg.app

    program = Program()
    import_helper = ImportHelper(program, GLOBAL_IDENTS)

    program.import_("__future__.annotations", True)
    program.import_("litestar.put", True)
    program.import_("sera.libs.digraph.ItemID", True)
    program.import_("sera.libs.digraph.LitestarContext", True)
    program.import_("sera.libs.digraph.NodeInput", True)
    program.import_(
        f"{app.logic.path}.digraph.graph",
        True,
    )
    program.import_(
        app.models.data.path
        + f".{collection.get_pymodule_name()}.Update{collection.name}",
        True,
    )

    # assuming the collection has only one class
    cls = collection.cls
    id_prop = assert_not_null(cls.get_id_property())
    id_type = id_prop.datatype.get_python_type().type
    collection_module = collection.get_pymodule_name()
    node_id = f"{collection_module}.update"

    update_controlled_props = [
        prop
        for prop in cls.properties.values()
        if prop.data.system_controlled is not None
        and prop.data.system_controlled.is_on_update_value_updated()
    ]
    if len(update_controlled_props) > 0:
        program.import_("sera.libs.api_helper.SingleAutoUSCP", True)

    # Check if any controlled property is the ID property itself
    # In this case, we have to reject the request if the ID doesn't match
    is_id_controlled = any(
        prop.name == id_prop.name for prop in update_controlled_props
    )
    if is_id_controlled:
        program.import_("litestar.exceptions.HTTPException", True)
        program.import_("litestar.status_codes", True)
        program.import_(
            "sera.libs.api_helper.SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY", True
        )
        program.import_("litestar.types.Scope", True)

    available_controlled_prop_vars: AvailableVars = {
        "user": PredefinedFn.item_getter(
            expr.ExprIdent("scope"),
            expr.ExprConstant("user"),
        )
    }

    func_name = "update"

    program.root(
        stmt.LineBreak(),
        stmt.PythonDecoratorStatement(
            expr.ExprFuncCall(
                expr.ExprIdent("put"),
                [
                    expr.ExprConstant("/{id:%s}" % id_type),
                ]
                + (
                    [
                        PredefinedFn.keyword_assignment(
                            "dto",
                            PredefinedFn.item_getter(
                                expr.ExprIdent("SingleAutoUSCP"),
                                expr.ExprIdent(f"Update{cls.name}"),
                            ),
                        )
                    ]
                    if len(update_controlled_props) > 0
                    else []
                ),
            )
        ),
        lambda ast10: ast10.func(
            func_name,
            [
                DeferredVar.simple(
                    "id",
                    expr.ExprIdent(id_type),
                ),
                DeferredVar.simple(
                    "data",
                    expr.ExprIdent(f"Update{cls.name}"),
                ),
                DeferredVar.simple(
                    "session",
                    import_helper.use("AsyncSession"),
                ),
                DeferredVar.simple("scope", expr.ExprIdent("Scope"))
                if is_id_controlled
                else None,
            ],
            return_type=expr.ExprIdent(id_prop.datatype.get_python_type().type),
            is_async=True,
        )(
            stmt.SingleExprStatement(expr.ExprConstant("Update an existing record")),
            *(
                [
                    stmt.LineBreak(),
                    lambda ast_if: ast_if.if_(
                        # if not scope["state"][SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY]:
                        expr.ExprNegation(
                            PredefinedFn.item_getter(
                                PredefinedFn.item_getter(
                                    expr.ExprIdent("scope"),
                                    expr.ExprConstant("state"),
                                ),
                                expr.ExprIdent(
                                    "SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY"
                                ),
                            )
                        )
                    )(
                        # if scope["user"].tenant_id != id:
                        lambda ast_inner_if: ast_inner_if.if_(
                            expr.ExprNotEqual(
                                get_controlled_property_value(
                                    assert_not_null(
                                        id_prop.data.system_controlled
                                    ).get_on_search_update_func(),
                                    available_controlled_prop_vars,
                                ),
                                expr.ExprIdent("id"),
                            )
                        )(
                            lambda ast_raise: ast_raise.raise_exception(
                                expr.StandardExceptionExpr(
                                    expr.ExprIdent("HTTPException"),
                                    [
                                        PredefinedFn.keyword_assignment(
                                            "status_code",
                                            expr.ExprIdent(
                                                "status_codes.HTTP_404_NOT_FOUND"
                                            ),
                                        ),
                                        PredefinedFn.keyword_assignment(
                                            "detail",
                                            expr.ExprIdent(
                                                'f"Record with id {id} not found"'
                                            ),
                                        ),
                                    ],
                                )
                            )
                        ),
                    ),
                    stmt.LineBreak(),
                ]
                if is_id_controlled
                else []
            ),
            # data.id = id
            stmt.SingleExprStatement(
                PredefinedFn.attr_setter(
                    expr.ExprIdent("data"),
                    expr.ExprIdent(id_prop.name),
                    expr.ExprIdent("id"),
                )
            ),
            # context = LitestarContext(session=session)
            lambda ast100: ast100.assign(
                DeferredVar.simple("context"),
                expr.ExprFuncCall(
                    expr.ExprIdent("LitestarContext"),
                    [
                        PredefinedFn.keyword_assignment(
                            "session", expr.ExprIdent("session")
                        ),
                    ],
                ),
            ),
            # input = NodeInput(id=ItemID("0"), args=data.to_db())
            lambda ast101: ast101.assign(
                DeferredVar.simple("input"),
                expr.ExprFuncCall(
                    expr.ExprIdent("NodeInput"),
                    [
                        PredefinedFn.keyword_assignment(
                            "id",
                            expr.ExprFuncCall(
                                expr.ExprIdent("ItemID"),
                                [expr.ExprConstant("0")],
                            ),
                        ),
                        PredefinedFn.keyword_assignment(
                            "args",
                            expr.ExprMethodCall(expr.ExprIdent("data"), "to_db", []),
                        ),
                    ],
                ),
            ),
            # result = await graph.execute_async({"<node_id>": input}, context)
            lambda ast102: ast102.assign(
                DeferredVar.simple("result"),
                expr.ExprAwait(
                    expr.ExprMethodCall(
                        expr.ExprIdent("graph"),
                        "execute_async",
                        [
                            PredefinedFn.dict(
                                [
                                    (
                                        expr.ExprConstant(node_id),
                                        expr.ExprIdent("input"),
                                    )
                                ]
                            ),
                            expr.ExprIdent("context"),
                        ],
                    )
                ),
            ),
            # return result["<node_id>"][0][0].value.id
            lambda ast103: ast103.return_(
                PredefinedFn.attr_getter(
                    PredefinedFn.attr_getter(
                        PredefinedFn.item_getter(
                            PredefinedFn.item_getter(
                                PredefinedFn.item_getter(
                                    expr.ExprIdent("result"),
                                    expr.ExprConstant(node_id),
                                ),
                                expr.ExprConstant(0),
                            ),
                            expr.ExprConstant(0),
                        ),
                        expr.ExprIdent("value"),
                    ),
                    expr.ExprIdent(id_prop.name),
                )
            ),
        ),
    )

    outmod = target_pkg.module("update")
    outmod.write(program)

    return outmod, func_name



from __future__ import annotations

from typing import Sequence

from sera.make.py_backend.logic.make_create_logic import make_python_create_logic
from sera.make.py_backend.logic.make_logic_graph import make_python_logic_graph
from sera.make.py_backend.logic.make_search_logic import make_python_search_logic
from sera.make.py_backend.logic.make_update_logic import make_python_update_logic
from sera.models import App, DataCollection


def make_python_logic(app: App, collections: Sequence[DataCollection]):
    """Make the basic structure for the logic."""
    app.logic.ensure_exists()

    for collection in collections:
        make_python_create_logic(collection, app.logic)
        make_python_update_logic(collection, app.logic)
        make_python_search_logic(collection, app.logic)

    # Generate digraph.py after all collection logic modules are created
    make_python_logic_graph(app, collections)



from __future__ import annotations

from codegen.models import (
    Program,
    expr,
    stmt,
)

from sera.models import (
    Package,
    Schema,
)
from sera.typing import ObjectPath


def make_python_enums(
    schema: Schema,
    target_pkg: Package,
    reference_objects: dict[str, ObjectPath],
):
    """Make enums defined in the schema.

    Args:
        schema: The schema to generate the classes from.
        target_pkg: The package to write the enums to.
        reference_objects: A dictionary of objects to their references (e.g., the ones that are defined outside and used as referenced such as Tenant).
    """
    for enum in schema.enums.values():
        if enum.name in reference_objects:
            # skip enums that are defined in different apps
            continue

        program = Program()
        program.import_("__future__.annotations", True)
        program.import_("enum.Enum", True)

        enum_values = []
        for value in enum.values.values():
            enum_values.append(
                stmt.DefClassVarStatement(
                    name=value.name,
                    type=None,
                    value=expr.ExprConstant(value.value),
                )
            )

        program.root(
            stmt.LineBreak(),
            lambda ast: ast.class_(
                enum.name,
                (
                    [expr.ExprIdent("str")]
                    if enum.is_str_enum()
                    else [expr.ExprIdent("int")]
                )
                + [expr.ExprIdent("Enum")],
            )(*enum_values),
        )

        target_pkg.module(enum.get_pymodule_name()).write(program)



from __future__ import annotations

from codegen.models import PredefinedFn, Program, expr, stmt

from sera.make.py_backend.misc import get_python_property_name
from sera.misc import assert_not_null, to_camel_case
from sera.models import Class, DataProperty, ObjectProperty, Package, Schema


def make_query(schema: Schema, cls: Class, pkg: Package):
    """Make query processor and query schema.

    Args:
        schema: The overall schema of the application, which contains all classes & enums
        cls: The class that we want to generate the query processor and schema
        pkg: The output package (directory) for the class in the `@.models` package. For example, if the
            class is `User`, then the package would be `src/models/user`.

    Returns:
        This function do not return anything as it writes the query helper directly to a file.
    """
    if not cls.is_public:
        # skip classes that are not public
        return

    outmod = pkg.module(cls.get_tsmodule_name() + "-query")

    program = Program()
    program.import_(
        f"@.models.{pkg.dir.name}.{cls.get_tsmodule_name()}.{cls.name}", True
    )
    program.import_(f"sera-db.QueryProcessor", True)
    program.import_(f"sera-db.Query", True)

    query_args = []
    for prop in cls.properties.values():
        pypropname = prop.name
        tspropname = to_camel_case(prop.name)

        if isinstance(prop, ObjectProperty) and prop.target.db is not None:
            tspropname = tspropname + "Id"
            pypropname = get_python_property_name(prop)

        if tspropname != pypropname:
            query_args.append(
                (
                    expr.ExprIdent(tspropname),
                    expr.ExprConstant(pypropname),
                )
            )

    query_condition_args = []
    for prop in cls.properties.values():
        if prop.db is None or prop.data.is_private:
            # This property is not stored in the database or it's private, so we skip it
            continue
        if (
            isinstance(prop, DataProperty)
            and prop.db is not None
            and not prop.db.is_indexed
        ):
            # This property is not indexed, so we skip it
            continue
        if isinstance(prop, ObjectProperty) and prop.target.db is None:
            # TODO: Implement this! This property is an embedded object property, we need to figure out
            # which necessary properties are queryable and add them to the field names
            continue

        tspropname = to_camel_case(prop.name)
        if isinstance(prop, ObjectProperty) and prop.target.db is not None:
            # This property is an object property stored in the database, "Id" is added to the property name
            tspropname = tspropname + "Id"

        if isinstance(prop, DataProperty):
            tstype = prop.datatype.get_typescript_type()
        else:
            assert isinstance(prop, ObjectProperty)
            tstype = assert_not_null(
                prop.target.get_id_property()
            ).datatype.get_typescript_type()

        for dep in tstype.deps:
            program.import_(dep, is_import_attr=True)

        query_ops = []

        if tstype.type == "string":
            query_ops.append(('"fuzzy"', tstype.type))
        elif tstype.type == "number":
            if (
                isinstance(prop, DataProperty)
                and prop.db is not None
                and prop.db.is_primary_key
            ) or (isinstance(prop, ObjectProperty) and prop.target.db is not None):
                # primary key or foreign key, we only support a limited set of operations
                query_ops.append(('"eq" | "ne"', tstype.type))
            else:
                query_ops.append(
                    ('"eq" | "ne" | "lt" | "lte" | "gt" | "gte"', tstype.type)
                )
                query_ops.append(('"bti"', "[number, number]"))
        elif tstype.is_enum_type():
            query_ops.append(('"eq" | "ne"', tstype.type))
        elif tstype.type == "Date":
            # for date type, we use iso string as the value
            query_ops.append(('"lte" | "gte"', "string"))
            query_ops.append(('"bti"', "[string, string]"))
        else:
            raise NotImplementedError(tstype.type)

        query_condition_args.append(
            (
                expr.ExprIdent(tspropname + "?"),
                expr.ExprRawTypescript(
                    " | ".join(
                        [
                            PredefinedFn.dict(
                                [
                                    (expr.ExprIdent("op"), expr.ExprIdent(op)),
                                    (expr.ExprIdent("value"), expr.ExprIdent(value)),
                                ]
                            ).to_typescript()
                            for op, value in query_ops
                        ]
                    )
                ),
            )
        )

    program.root(
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export const query = "
            + expr.ExprNewInstance(
                expr.ExprIdent(f"QueryProcessor<{cls.name}>"),
                [
                    PredefinedFn.dict(query_args),
                ],
            ).to_typescript()
            + ";",
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export type {cls.name}Query = Query<{cls.name}, "
            + PredefinedFn.dict(query_condition_args).to_typescript()
            + ">;"
        ),
    )

    outmod.write(program)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from dirty_equals import Contains, IsPartialDict
from litestar import Litestar
from litestar.testing import TestClient


@dataclass
class SearchExample:
    """A search test example with query and expected results.

    Attributes:
        query: The search query to send to the API.
        expected_records: A list of expected records, where each record is a
            sub-dictionary that should match at least one record in the results.
        unexpected_records: A list of records that should NOT appear in the
            search results. Each record is a sub-dictionary that must not match
            any record in the results.
    """

    query: dict
    expected_records: list[dict] = field(default_factory=list)
    unexpected_records: list[dict] = field(default_factory=list)


def get_record_by_id(
    client: TestClient[Litestar],
    base_url: str,
    record_id: int | str,
) -> dict:
    """Fetch a record by ID and return the extracted record data.

    The API response is usually in the format { "ModelName": [ record ] }.
    This function extracts and returns the first record.
    """
    resp = client.get(f"{base_url}/{record_id}")
    assert resp.status_code == 200, (
        f"Failed to get record {record_id}: {resp.status_code} - {resp.text}"
    )
    # The response is usually in the format { "ModelName": [ record ] }
    # We take the first value's first element.
    return list(resp.json().values())[0][0]


def test_has(
    client: TestClient[Litestar],
    base_url: str,
    exist_records: list[int | str],
    non_exist_records: list[int | str],
) -> None:
    for record in exist_records:
        resp = client.head(f"{base_url}/{record}")
        assert resp.status_code == 204, (
            f"Record {record} should exist but got {resp.status_code}"
        )

    for record in non_exist_records:
        resp = client.head(f"{base_url}/{record}")
        assert resp.status_code == 404, (
            f"Record {record} should not exist but got {resp.status_code}"
        )


def test_get_by_id(
    client: TestClient[Litestar],
    base_url: str,
    exist_records: dict[int | str, dict],
    non_exist_records: list[int | str],
) -> None:
    for record, data in exist_records.items():
        resp = client.get(f"{base_url}/{record}")
        assert resp.status_code == 200, (
            f"Record {record} should exist but got {resp.status_code}"
        )
        assert resp.json() == data, (resp.json(), data)

    for record in non_exist_records:
        resp = client.get(f"{base_url}/{record}")
        assert resp.status_code == 404, (
            f"Record {record} should not exist but got {resp.status_code}"
        )


def test_create(
    client: TestClient[Litestar],
    base_url: str,
    create_examples: list[dict],
    compare_properties: Optional[list[str]] = None,
) -> list[int]:
    """Test creating records and return the created IDs."""
    created_ids = []
    for data in create_examples:
        resp = client.post(base_url, json=data)
        assert resp.status_code == 201, (
            f"Failed to create record: {resp.status_code} - {resp.text}"
        )
        assert isinstance(resp.json(), int), f"Expected ID as int, got {resp.json()}"
        id = resp.json()
        created_ids.append(id)

        if compare_properties is not None:
            record = get_record_by_id(client, base_url, id)
            expected = {prop: data[prop] for prop in compare_properties}
            assert record == IsPartialDict(expected), (
                f"Created record does not match expected properties. "
                f"Expected {expected}, got {record}"
            )

    return created_ids


def test_update(
    client: TestClient[Litestar],
    base_url: str,
    update_examples: dict[int | str, dict],
    compare_properties: Optional[list[str]] = None,
) -> None:
    for record_id, data in update_examples.items():
        resp = client.put(f"{base_url}/{record_id}", json=data)
        assert resp.status_code == 200, (
            f"Failed to update record {record_id}: {resp.status_code} - {resp.text}"
        )
        assert resp.json() == record_id, (
            f"Expected returned ID {record_id}, got {resp.json()}"
        )

        if compare_properties is not None:
            record = get_record_by_id(client, base_url, record_id)
            expected = {prop: data[prop] for prop in compare_properties}
            assert record == IsPartialDict(expected), (
                f"Updated record does not match expected properties. "
                f"Expected {expected}, got {record}"
            )


def test_search(
    client: TestClient[Litestar],
    base_url: str,
    search_examples: list[SearchExample],
) -> None:
    """Test search functionality.

    Args:
        client: The test client.
        base_url: The base URL for the API.
        search_examples: A list of SearchExample objects, each containing a query
            and expected records to verify in the search results.

    Note:
        Uses dirty-equals for ergonomic assertions. Each expected record is
        matched as a partial dictionary against the search results.
    """
    for example in search_examples:
        resp = client.post(f"{base_url}/q", json=example.query)
        assert resp.status_code == 200, (
            f"Failed to search: {resp.status_code} - {resp.text}"
        )
        result = resp.json()
        assert isinstance(result, dict), f"Expected dict response, got {type(result)}"

        # Extract records from the response (format: { "ModelName": [ records ] })
        records = list(result.values())[0] if result else []

        # Use dirty-equals Contains with IsPartialDict for ergonomic assertions
        for expected in example.expected_records:
            assert records == Contains(IsPartialDict(expected)), (
                f"Expected record {expected} not found in search results. "
                f"Got records: {records}"
            )

        # Verify unexpected records are NOT in the results
        for unexpected in example.unexpected_records:
            assert records != Contains(IsPartialDict(unexpected)), (
                f"Unexpected record {unexpected} was found in search results. "
                f"Got records: {records}"
            )



from __future__ import annotations

from pathlib import Path
from typing import Sequence

import serde.yaml

from sera.models._property import (
    DataProperty,
    ForeignKeyOnDelete,
    ForeignKeyOnUpdate,
    WeakRef,
)
from sera.models._schema import Schema
from sera.models.parse.parse_class import parse_class_without_prop
from sera.models.parse.parse_enum import parse_enum
from sera.models.parse.parse_event import parse_data_event
from sera.models.parse.parse_property import parse_property


def parse_schema(name: str, files: Sequence[Path | str]) -> Schema:
    schema = Schema(name=name, classes={}, enums={})

    # parse all classes
    raw_defs = {}
    for file in files:
        for k, v in serde.yaml.deser(file).items():
            if k.startswith("enum:"):
                schema.enums[k[5:]] = parse_enum(schema, k[5:], v)
                continue
            cdef = parse_class_without_prop(schema, k, v)
            assert k not in schema.classes
            schema.classes[k] = cdef
            raw_defs[k] = v

    # now parse properties of the classes
    for clsname, v in raw_defs.items():
        cdef = schema.classes[clsname]
        for propname, prop in (v["props"] or {}).items():
            assert propname not in cdef.properties
            parsed_prop = parse_property(schema, cdef, propname, prop)
            cdef.properties[propname] = parsed_prop

    resolve_weakrefs(schema, raw_defs)

    # parse events after all properties are defined
    all_props = {
        f"{cls.name}.{prop.name}": prop
        for cls in schema.classes.values()
        for prop in cls.properties.values()
    }

    for clsname, v in raw_defs.items():
        cdef = schema.classes[clsname]
        all_props = dict(all_props, **cdef.properties)
        for event_data in v.get("events", []):
            event = parse_data_event(event_data, all_props)
            cdef.events.append(event)

    return schema


def resolve_weakrefs(schema: Schema, raw_defs: dict[str, dict]) -> None:
    for clsname, raw_def in raw_defs.items():
        cls = schema.classes[clsname]
        for prop_name, prop_def in (raw_def.get("props") or {}).items():
            if "weakref" not in prop_def:
                continue

            prop = cls.properties[prop_name]
            if not isinstance(prop, DataProperty):
                raise ValueError(
                    f"Weakref is only supported for data properties: {cls.name}.{prop.name}"
                )
            if cls.db is None:
                raise ValueError(
                    f"Weakref requires database class: {cls.name}.{prop.name}"
                )
            if prop.db is None:
                raise ValueError(
                    f"Weakref requires database property: {cls.name}.{prop.name}"
                )
            if prop.db.foreign_key is not None:
                raise ValueError(
                    f"Weakref cannot be combined with db.foreign_key: {cls.name}.{prop.name}"
                )
            weakref_data = prop_def["weakref"]
            if not isinstance(weakref_data, dict):
                raise ValueError(
                    f"Weakref for {cls.name}.{prop.name} must be a mapping"
                )
            target = weakref_data.get("target")
            if not isinstance(target, str) or target.strip() == "":
                raise ValueError(
                    f"Weakref target for {cls.name}.{prop.name} must be a non-empty string"
                )
            if "." not in target:
                raise ValueError(
                    f"Weakref target must be in format Class.prop: {cls.name}.{prop.name}"
                )
            target_cls_name, target_prop_name = target.split(".", 1)
            target_cls = schema.classes.get(target_cls_name)
            if target_cls is None:
                raise ValueError(
                    f"Weakref target class not found: {target_cls_name} for {cls.name}.{prop.name}"
                )
            target_prop = target_cls.properties.get(target_prop_name)
            if target_prop is None:
                raise ValueError(
                    f"Weakref target property not found: {target_cls_name}.{target_prop_name}"
                )
            if not isinstance(target_prop, DataProperty):
                raise ValueError(
                    f"Weakref target must be a data property: {target_cls_name}.{target_prop_name}"
                )
            if target_prop.db is None:
                raise ValueError(
                    f"Weakref target property must be stored in DB: {target_cls_name}.{target_prop_name}"
                )
            if prop.datatype != target_prop.datatype:
                raise ValueError(
                    "Weakref data type mismatch: "
                    f"{cls.name}.{prop.name} ({prop.datatype}) vs "
                    f"{target_cls_name}.{target_prop_name} ({target_prop.datatype})"
                )
            if (
                ForeignKeyOnDelete(weakref_data.get("on_delete", "restrict")).value
                == "set null"
                and not prop.is_optional
            ):
                raise ValueError(
                    f"Weakref on_delete set null requires optional property: {cls.name}.{prop.name}"
                )

            prop.weakref = WeakRef(
                target_class=target_cls,
                target_prop=target_prop,
                on_delete=ForeignKeyOnDelete(weakref_data.get("on_delete", "restrict")),
                on_update=ForeignKeyOnUpdate(weakref_data.get("on_update", "restrict")),
            )



from __future__ import annotations

import os
from pathlib import Path
from pydoc import doc
from typing import Annotated

from codegen.models import DeferredVar, PredefinedFn, Program, expr, stmt

from sera.make.make_typescript_model import make_typescript_data_model
from sera.make.py_backend import (
    make_python_api,
    make_python_data_model,
    make_python_enums,
    make_python_logic,
    make_python_relational_model,
)
from sera.make.ts_frontend.make_enums import make_typescript_enums
from sera.misc import CacheCleaner, FileWriter, check_clean_repo_or_token
from sera.models import App, DataCollection, parse_schema
from sera.typing import Language


def make_config(app: App):
    """Make the configuration for the application."""
    program = Program()
    program.import_("__future__.annotations", True)
    program.import_("os", False)
    program.import_("serde.yaml", False)
    program.import_("pathlib.Path", True)
    program.import_("sera.models.parse_schema", True)

    program.root(
        lambda ast: ast.if_(
            PredefinedFn.not_has_item(
                expr.ExprIdent("os.environ"), expr.ExprConstant("CFG_FILE")
            )
        )(
            lambda ast01: ast01.assign(
                DeferredVar.simple("CFG_FILE"),
                expr.ExprRawPython("Path(__file__).parent.parent / 'config.yml'"),
            ),
        ),
        lambda ast: ast.else_()(
            lambda ast01: ast01.assign(
                DeferredVar.simple("CFG_FILE"),
                expr.ExprRawPython('Path(os.environ["CFG_FILE"])'),
            ),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("cfg"),
            expr.ExprFuncCall(
                expr.ExprIdent("serde.yaml.deser"), [expr.ExprIdent("CFG_FILE")]
            ),
        ),
        stmt.LineBreak(),
        lambda ast: ast.assign(
            DeferredVar.simple("DB_CONNECTION"),
            expr.ExprIdent("cfg['db']['connection']"),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("DB_DEBUG"),
            expr.ExprIdent('os.environ.get("DB_DEBUG", "0") == "1"'),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("API_DEBUG"),
            expr.ExprIdent('os.environ.get("API_DEBUG", "0") == "1"'),
        ),
        stmt.LineBreak(),
        lambda ast: ast.assign(
            DeferredVar.simple("PKG_DIR"),
            expr.ExprRawPython("Path(__file__).parent"),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("SCHEMA_DIR"),
            expr.ExprRawPython("PKG_DIR.parent / 'schema'"),
        ),
        lambda ast: ast.assign(
            DeferredVar.simple("schema"),
            expr.ExprFuncCall(
                expr.ExprIdent("parse_schema"),
                [
                    expr.ExprConstant(app.name),
                    PredefinedFn.list(
                        [
                            expr.ExprDivision(
                                expr.ExprIdent("SCHEMA_DIR"),
                                expr.ExprConstant(
                                    os.path.relpath(
                                        path.absolute(), app.root.dir.parent / "schema"
                                    )
                                ),
                            )
                            for path in app.schema_files
                        ]
                    ),
                ],
            ),
        ),
    )

    app.config.write(program)


def make_app(
    app_dir: Annotated[
        Path,
        doc("Directory of the generated application"),
    ],
    schema_files: Annotated[
        list[Path],
        doc("YAML schema files. Multiple files are merged automatically"),
    ],
    api_collections: Annotated[
        list[str],
        doc("API collections to generate."),
    ],
    language: Annotated[
        Language,
        doc(
            "Language of the generated application. Currently only Python is supported."
        ),
    ] = Language.Python,
    referenced_schema: Annotated[
        list[str],
        doc(
            "Classes in the schema that are defined in different modules, used as references and thus should not be generated."
        ),
    ] = [],
    allow_dirty_repo: Annotated[
        bool,
        doc(
            "Allow code generation in a repo with uncommitted changes (by default, clean repo is enforced)"
        ),
    ] = False,
    session_token: Annotated[
        str | None,
        doc(
            "Token filename for parallel runs. If the token file exists, version control checks are skipped."
        ),
    ] = None,
):
    # Perform version control safety check before code generation
    check_clean_repo_or_token(app_dir, session_token, allow_dirty_repo)

    schema = parse_schema(app_dir.name, schema_files)

    app = App(app_dir.name, app_dir, schema_files, language)

    if language == Language.Python:
        # generate application configuration
        make_config(app)

        # generate models from schema
        # TODO: fix me, this is a hack to make the code work for referenced classes
        referenced_classes = {
            path.rsplit(".", 1)[1]: (parts := path.rsplit(".", 1))[0]
            + ".data."
            + parts[1]
            for path in referenced_schema
        }

        make_python_enums(schema, app.models.pkg("enums"), referenced_classes)
        make_python_data_model(schema, app.models.pkg("data"), referenced_classes)
        referenced_classes = {
            path.rsplit(".", 1)[1]: (parts := path.rsplit(".", 1))[0]
            + ".db."
            + parts[1]
            for path in referenced_schema
        }
        make_python_relational_model(
            schema,
            app.models.pkg("db"),
            app.models.pkg("data"),
            referenced_classes,
        )

        collections = [
            DataCollection(schema.classes[cname]) for cname in api_collections
        ]

        # generate API
        make_python_api(app, collections)

        # generate logic
        make_python_logic(app, collections)
    elif language == Language.Typescript:
        make_typescript_enums(schema, app.models)
        make_typescript_data_model(schema, app.models)

    FileWriter.get_instance().process()
    CacheCleaner(app_dir).cleanup()
    return app


if __name__ == "__main__":
    make_app(
        Path("/Volumes/research/workspace/libs/sera/tests/resources/myapp"),
        [
            Path(
                "/Volumes/research/workspace/libs/sera/tests/resources/schema/product.yml"
            )
        ],
        ["Product", "Category"],
    )



from __future__ import annotations

from sera.models import DataProperty, ObjectProperty


def get_python_property_name(prop: DataProperty | ObjectProperty) -> str:
    """Get property name of a property in Python model"""
    if isinstance(prop, ObjectProperty) and prop.target.db is not None:
        # the property value is a foreign key to another table, we should add _id to
        # the end of the property name
        return prop.name + "_id"
    else:
        return prop.name



from __future__ import annotations

from dataclasses import dataclass, field

from sera.models._expression import Expr


@dataclass
class EventCondition:
    """Represents a condition in Conjunctive Normal Form (CNF).

    CNF is represented as a list of OR clauses, where all must be true (AND).
    Each OR clause is a list of EventClause where at least one must be true.

    Example CNF:
    - [[A], [B, C]] represents: A AND (B OR C)
    - [[A, B]] represents: (A OR B)
    - [[A], [B], [C]] represents: A AND B AND C
    """

    # List of OR clauses (each OR clause is a list of EventClause)
    # All OR clauses must evaluate to true (AND relationship)
    clauses: list[list[Expr]] = field(default_factory=list)



from __future__ import annotations

import inspect
from typing import Any, Callable, Coroutine, Generic, TypeVar, get_args, get_type_hints

from graph.interface import BaseNode

from sera.libs.digraph._context import Context
from sera.libs.digraph._execution_summary import ExecutionSummary
from sera.libs.digraph._node_input import NodeInput
from sera.libs.digraph._node_output import NodeOutput
from sera.libs.digraph._types import NodeID

I = TypeVar("I")
O = TypeVar("O")
C = TypeVar("C", bound=Context)

# Type alias for sync compute function signature
SyncComputeFn = Callable[[NodeInput[I], C], tuple[NodeOutput[O], ExecutionSummary]]

# Type alias for async compute function signature
AsyncComputeFn = Callable[
    [NodeInput[I], C],
    Coroutine[None, None, tuple[NodeOutput[O], ExecutionSummary]],
]

# Combined type for both sync and async compute functions
ComputeFn = SyncComputeFn[I, C, O] | AsyncComputeFn[I, C, O]


class Node(BaseNode[NodeID], Generic[I, C, O]):
    """A node in the computing graph.

    The compute function takes:
    - input: NodeInput[I] containing the item ID and arguments
    - context: Context providing access to previous outputs and global state

    And returns:
    - A tuple of (NodeOutput[O], ExecutionSummary)

    Example:
        >>> def process_order(input: NodeInput[Order], context: Context) -> tuple[NodeOutput[ProcessedOrder], ExecutionSummary]:
        ...     order = input.args
        ...     result = ProcessedOrder(...)
        ...     return NodeOutput(value=result), CreateSummary()
        ...
        >>> node = Node(id="process_order", compute=process_order)
    """

    def __init__(self, id: NodeID, compute: ComputeFn[I, C, O]):
        super().__init__(id)
        self.compute = compute
        self.is_async = inspect.iscoroutinefunction(compute)
        self.input_type, self.output_type = _extract_input_output_types(compute)


def _extract_input_output_types(
    compute: ComputeFn[Any, Any, Any],
) -> tuple[type, type]:
    """Extract input and output types from compute function signature.

    Args:
        compute: The compute function to extract types from.

    Returns:
        A tuple of (input_type, output_type).

    Raises:
        TypeError: If input or output types cannot be inferred from the function signature.
    """
    func_name = getattr(compute, "__name__", repr(compute))

    try:
        hints = get_type_hints(compute)
    except Exception as e:
        raise TypeError(
            f"Cannot get type hints for compute function '{func_name}': {e}"
        ) from e

    input_type = None
    output_type = None

    # Get the first parameter's type hint (should be NodeInput[T])
    params = list(inspect.signature(compute).parameters.values())
    if params and params[0].name in hints:
        node_input_hint = hints[params[0].name]
        # Extract T from NodeInput[T]
        args = get_args(node_input_hint)
        if args:
            input_type = args[0]

    if input_type is None:
        raise TypeError(
            f"Cannot infer input type for compute function '{func_name}'. "
            f"Ensure the first parameter is typed as NodeInput[T]."
        )

    # Get return type (should be tuple[NodeOutput[O], ExecutionSummary])
    if "return" in hints:
        return_hint = hints["return"]
        return_args = get_args(return_hint)
        if return_args and len(return_args) >= 1:
            # First element should be NodeOutput[O]
            node_output_hint = return_args[0]
            output_args = get_args(node_output_hint)
            if output_args:
                output_type = output_args[0]

    if output_type is None:
        raise TypeError(
            f"Cannot infer output type for compute function '{func_name}'. "
            f"Ensure the return type is typed as tuple[NodeOutput[O], ExecutionSummary]."
        )

    return input_type, output_type



from __future__ import annotations

from typing import Optional

from msgspec.json import decode, encode
from sqlalchemy import LargeBinary, TypeDecorator, update
from sqlalchemy import create_engine as sqlalchemy_create_engine
from sqlalchemy.ext.asyncio import create_async_engine as sqlalchemy_create_async_engine

from sera.typing import UNSET


class BaseORM:
    def get_update_query(self):
        q = update(self.__class__)
        args = {}
        for col in self.__table__.columns:  # type: ignore
            val = getattr(self, col.name)
            if val is UNSET:
                continue
            if col.primary_key:
                q = q.where(getattr(self.__class__, col.name) == val)
            args[col.name] = val

        return q.values(**args)

    def get_update_args(self):
        table = self.__table__  # type: ignore
        return {
            col.name: val
            for col in table.columns
            if (val := getattr(self, col.name)) is not UNSET
        }

    @classmethod
    def from_dict(cls, data: dict):
        raise NotImplementedError()


class DataclassType(TypeDecorator):
    """SqlAlchemy Type decorator to serialize dataclasses"""

    impl = LargeBinary
    cache_ok = True

    def __init__(self, cls):
        super().__init__()
        self.cls = cls

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return encode(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return decode(value, type=self.cls)


class ListDataclassType(TypeDecorator):
    """SqlAlchemy Type decorator to serialize list of dataclasses"""

    impl = LargeBinary
    cache_ok = True

    def __init__(self, cls):
        super().__init__()
        self.cls = cls

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return encode(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return decode(value, type=list[self.cls])


class DictDataclassType(TypeDecorator):
    """SqlAlchemy Type decorator to serialize mapping of dataclasses"""

    impl = LargeBinary
    cache_ok = True

    def __init__(self, cls):
        super().__init__()
        self.cls = cls

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return encode(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return decode(value, type=dict[str, self.cls])


def create_engine(
    dbconn: str,
    connect_args: Optional[dict] = None,
    echo: bool = False,
):
    if dbconn.startswith("sqlite"):
        connect_args = {"check_same_thread": False}
    else:
        connect_args = {}
    engine = sqlalchemy_create_engine(dbconn, connect_args=connect_args, echo=echo)
    return engine


def create_async_engine(
    dbconn: str,
    connect_args: Optional[dict] = None,
    echo: bool = False,
):
    if dbconn.startswith("sqlite"):
        connect_args = {"check_same_thread": False}
    else:
        connect_args = {}

    engine = sqlalchemy_create_async_engine(
        dbconn, connect_args=connect_args, echo=echo
    )
    return engine



from __future__ import annotations

from typing import Callable

from litestar.connection.base import UserT
from litestar.middleware import AbstractMiddleware
from litestar.types import ASGIApp, Message, Receive, Scope, Scopes, Send

SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY = "skip_uscp_1157"


class USCPMiddleware(AbstractMiddleware):
    """
    Middleware to update system-controlled properties in the request.

    This middleware updates the `created_at`, `updated_at`, and `deleted_at` properties
    of the request with the current timestamp. It is intended to be used in a Litestar
    application.
    """

    def __init__(
        self,
        app: ASGIApp,
        skip_update_system_controlled_props: Callable[[UserT], bool],
        exclude: str | list[str] | None = None,
        exclude_opt_key: str | None = None,
        scopes: Scopes | None = None,
    ) -> None:
        """Initialize the middleware.

        Args:
            app: The ``next`` ASGI app to call.
            exclude: A pattern or list of patterns to match against a request's path.
                If a match is found, the middleware will be skipped.
            exclude_opt_key: An identifier that is set in the route handler
                ``opt`` key which allows skipping the middleware.
            scopes: ASGI scope types, should be a set including
                either or both 'ScopeType.HTTP' and 'ScopeType.WEBSOCKET'.
        """
        super().__init__(app, exclude, exclude_opt_key, scopes)
        self.skip_update_system_controlled_props = skip_update_system_controlled_props

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        user = scope["user"]
        scope["state"][SKIP_UPDATE_SYSTEM_CONTROLLED_PROPS_KEY] = (
            self.skip_update_system_controlled_props(user)
        )
        await self.app(scope, receive, send)


def get_scp_from_user(user: UserT, fields: dict[str, str]) -> dict:
    """Get system-controlled properties from the user.

    Args:
        user: The user object.
        fields: A list of fields to include in the system-controlled properties.

    Returns:
        A dictionary containing the system-controlled properties.
    """
    return {
        data_field: getattr(user, db_field) for data_field, db_field in fields.items()
    }



from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, TypeGuard, TypeVar, Union

import msgspec


class doc(str):
    """A docstring for a type. Typically used in Annotated"""


T = TypeVar("T")
FieldName = Annotated[str, doc("field name of a class")]
ObjectPath = Annotated[
    str, doc("path of an object (e.g., can be function, class, etc.)")
]


class Language(str, Enum):
    Python = "python"
    Typescript = "typescript"


# re-export msgspec.UnsetType & msgspec.UNSET, so that we are consistent with ORM & data modules
UnsetType = msgspec.UnsetType
UNSET: Any = msgspec.UNSET


def is_set(value: Union[T, UnsetType]) -> TypeGuard[T]:
    """Typeguard to check if a value is set (not UNSET)"""
    return value is not UNSET


# Global identifiers for codegen
GLOBAL_IDENTS = {
    "AsyncSession": "sqlalchemy.ext.asyncio.AsyncSession",
    "ASGIConnection": "litestar.connection.ASGIConnection",
    "UNSET": "sera.typing.UNSET",
    "ForeignKey": "sqlalchemy.ForeignKey",
    "Optional": "typing.Optional",
    "text": "sqlalchemy.text",
}



"""Expression classes for building complex expressions in SERA models.

This module provides a hierarchy of expression classes that can be used to represent
various types of expressions including constants, property references, boolean operations,
function calls, and arithmetic operations.
"""

from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from sera.models._property import Property


class Expr(ABC):
    """Base class for all expression types."""


@dataclass
class ConstantExpr(Expr):
    """Expression representing a constant value.

    Attributes:
        value: The constant value (can be str, int, float, bool, None, etc.)
    """

    value: Optional[str | int | float | bool | tuple[str, ...]]


@dataclass
class AttrGetterExpr(Expr):
    """Expression representing a property reference.

    Attributes:
        property: The Property instance being referenced
    """

    property: Property


class BoolOp(str, Enum):
    """Boolean operator types."""

    AND = "and"
    OR = "or"
    NOT = "not"


@dataclass
class LogicalExpr(Expr):
    """Expression representing a boolean operation.

    Attributes:
        operator: The boolean operator (AND, OR, NOT)
        operands: List of operand expressions
    """

    operator: BoolOp
    operands: list[Expr]

    def __post_init__(self):
        """Validate operands based on operator type."""
        if self.operator == BoolOp.NOT and len(self.operands) != 1:
            raise ValueError("NOT operator requires exactly one operand")
        if self.operator in (BoolOp.AND, BoolOp.OR):
            if len(self.operands) < 2:
                raise ValueError(
                    f"{self.operator.value.upper()} operator requires at least two operands"
                )


class ArithmeticOp(str, Enum):
    """Arithmetic operator types."""

    ADD = "+"
    SUBTRACT = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    MODULO = "%"
    POWER = "**"


# Comparison operators - common enough to warrant their own class
class ComparisonOp(str, Enum):
    """Comparison operator types."""

    EQ = "=="
    NEQ = "!="
    GT = ">"
    GTE = ">="
    LT = "<"
    LTE = "<="
    IN = "in"
    NOT_IN = "not in"
    IS = "is"
    IS_NOT = "is not"


@dataclass
class BinaryExpr(Expr):
    """Expression representing a binary operation (arithmetic or comparison).

    Attributes:
        operator: The binary operator (arithmetic or comparison)
        left: Left operand expression
        right: Right operand expression
    """

    operator: ArithmeticOp | ComparisonOp
    left: Expr
    right: Expr


@dataclass
class FuncCallExpr(Expr):
    """Expression representing a function call.

    Attributes:
        func_name: The name of the function to call
        args: List of argument expressions
        kwargs: Optional keyword arguments as expressions
    """

    func_name: str
    args: list[Expr]
    kwargs: dict[str, Expr] | None = None


__all__ = [
    "Expr",
    "ConstantExpr",
    "AttrGetterExpr",
    "LogicalExpr",
    "BoolOp",
    "BinaryExpr",
    "ArithmeticOp",
    "FuncCallExpr",
    "ComparisonOp",
]



from __future__ import annotations


from codegen.models import PredefinedFn, Program, expr, stmt

from sera.models import (
    Enum,
    Package,
    Schema,
)


def make_typescript_enums(schema: Schema, target_pkg: Package):
    """Make typescript enum for the schema"""
    enum_pkg = target_pkg.pkg("enums")

    for enum in schema.enums.values():
        make_enum(enum, enum_pkg)

    program = Program()
    for enum in schema.enums.values():
        program.import_(f"@.models.enums.{enum.get_tsmodule_name()}.{enum.name}", True)
        program.import_(
            f"@.models.enums.{enum.get_tsmodule_name()}.{enum.name}Label", True
        )
        program.import_(
            f"@.models.enums.{enum.get_tsmodule_name()}.{enum.name}Description", True
        )

    program.root(
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            "export { "
            + ", ".join([enum.name for enum in schema.enums.values()])
            + ","
            + ", ".join([enum.name + "Label" for enum in schema.enums.values()])
            + ","
            + ", ".join([enum.name + "Description" for enum in schema.enums.values()])
            + "};"
        ),
    )
    enum_pkg.module("index").write(program)


def make_enum(enum: Enum, pkg: Package):
    program = Program()
    program.root(
        stmt.LineBreak(),
        lambda ast: ast.class_like("enum", enum.name)(
            *[
                stmt.DefEnumValueStatement(
                    name=value.name,
                    value=expr.ExprConstant(value.value),
                )
                for value in enum.values.values()
            ]
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export const {enum.name}Label = "
            + PredefinedFn.dict(
                [
                    (
                        expr.ExprConstant(value.value),
                        expr.ExprConstant(value.label.to_dict()),
                    )
                    for value in enum.values.values()
                ]
            ).to_typescript()
        ),
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            f"export const {enum.name}Description = "
            + PredefinedFn.dict(
                [
                    (
                        expr.ExprConstant(value.value),
                        expr.ExprConstant(value.description.to_dict()),
                    )
                    for value in enum.values.values()
                ]
            ).to_typescript()
        ),
    )
    pkg.module(enum.get_tsmodule_name()).write(program)



from __future__ import annotations

from typing import Any, Callable

from graph.interface import BaseEdge

from sera.libs.digraph._node_output import NodeOutput
from sera.libs.digraph._types import NodeID


class Edge(BaseEdge[NodeID, int]):
    """An edge connecting two nodes in the graph.

    The filter function determines whether to propagate execution to the target node.
    If None (default), the edge always propagates.

    Args:
        source: The NodeID of the source node.
        target: The NodeID of the target node.
        filter_fn: Optional function that takes a NodeOutput and returns True
                   if execution should propagate to the target node.

    Example:
        >>> # Always propagate (default)
        >>> edge = Edge(source=NodeID("a"), target=NodeID("b"))
        >>>
        >>> # Only propagate if output value is positive
        >>> edge = Edge(
        ...     source=NodeID("a"),
        ...     target=NodeID("b"),
        ...     filter_fn=lambda output: output.value > 0
        ... )
    """

    def __init__(
        self,
        source: NodeID,
        target: NodeID,
        filter_fn: Callable[[NodeOutput[Any]], bool] | None = None,
        id: int = -1,
    ):
        # key=0 since we don't allow parallel edges
        super().__init__(id, source, target, key=0)
        self.filter_fn = filter_fn

    def should_propagate(self, output: NodeOutput[Any]) -> bool:
        """Check if execution should propagate through this edge.

        Args:
            output: The output from the source node.

        Returns:
            True if execution should propagate to the target, False otherwise.
        """
        if self.filter_fn is None:
            return True
        return self.filter_fn(output)
